/**
 * Memory Analyser  version 1.00
 * Copyright (c) 2018, Tycho Veltmeijer.
 * All rights reserved.
**/

#include "Filter_Firefox.h"

/*Firefox Filter contains 3 filters. Two of them search for URLS and one for Cookies, the filters are described in the dutch documentation: "3. Firefox".
Filter 1: URL's, described in "3.1 URL's",  "Methode 1"
Filter 2: URL's, described in "3.1 URL's",  "Methode 2"
Filter 3: Cookies, described in "3.2 Cookies" */


/*--------------------------------------------------------Filter 1--------------------------------------------------------*/
DUMP_SESSION* InitiateFirefoxFilter1(char* location, char* argument, int analyse_type) {
	return CreateUNICODEMemoryDump(location, "Firefox_URL_Search_Methode1.txt", 0, 
		(argument==(char*) -1 ? DS_WRITE_LOCATION_STRING : strtol(argument, NULL, 10))
	);
}

/*Search for UNICODE URL's  (using signature 0x00000001)*/
int ApplyFirefoxFilter1(DUMP_SESSION* dmp, HANDLE process, MEMORY_BASIC_INFORMATION* memInfo, char* buff, SIZE_T overflowSize) {
	char signature[4]={0x01, 0x00, 0x00, 0x00}; //'signature'
	char* moveBuff=buff; //cur search position 
	wchar_t* moveBuff2; //back-up of movebuff
	int writtenBytes; //Returnvalue of Writefile()
	int readSize=0;  //temp parameter with size of string
	int isOkayURL=0; //Boolean: true when uRL, false when not.
	int foundDot=0; //Boolean, true when dot is found in string, false if not

	if(memInfo->State!=MEM_FREE && memInfo->State!=MEM_RESERVE) {
		for(; moveBuff<(buff + memInfo->RegionSize); moveBuff++) {
			if(memcmp(moveBuff, signature, 4) == 0) { //found signature?
				moveBuff+=4; //skip signature
				readSize = (int) *moveBuff; //read length of possible url
				moveBuff+=4; //skip size

				/*set booleans on false*/
				isOkayURL = 0; 
				foundDot=0;
				
				//check if the size is indeed correct and the URL doesn't start with "chrome://"
				if(wcslen((wchar_t*)moveBuff)*2==readSize-2 
					&& readSize!=0 
					&& memcmp(moveBuff, L"chrome://", 9*2) != 0) 
					isOkayURL = 1;

				moveBuff2=(wchar_t*) moveBuff;

				/*does the domain meets the requirments:
				- letters a - z
				- numbers 0-9
				- '-' or '.' */
				while(*moveBuff2 != 0) {
					if((*moveBuff2 < 'a' || *moveBuff2 > 'z') 
					&& (*moveBuff2 < '0' || *moveBuff2 > '9') 
					&& *moveBuff2 != '-'
					&& *moveBuff2 != '.') {
						if(*moveBuff2 == ':' && *(moveBuff2+1) == '/' && *(moveBuff2+2) == '/') {//"://" is een protocol
							moveBuff2+=2;
						} else {
							if(*moveBuff2 == '/') //This is where the domain stops and the URL starts
								break;
							isOkayURL = 0;
							break;
						}
					} else {
						if(*moveBuff2 == '.')
							foundDot=1;
					}

					moveBuff2++;
				}

				if(foundDot==0)
					isOkayURL = 0;

				/*Check if the rest of the URL meets the requirments:
				- no space allowed (will be %20)
				- no new line (0D of 0A)
				*/
				while(*moveBuff2 != 0) {
					if(*moveBuff2 == ' ' || *moveBuff2 == 0x0D || *moveBuff2 == 0x0A) {
						isOkayURL = 0;
						break;
					}
					moveBuff2++;
				}

				/*We found a URL, write it to the file.*/
				if(isOkayURL) {
					if(dmp->flags&DS_WRITE_LOCATION_STRING) {
						printf("Address: %p: ",(memlong) memInfo->BaseAddress + (moveBuff - buff));
						wsprintfW((LPWSTR) dmp->string_location, L"Address: %p, \0", (memlong) memInfo->BaseAddress + (moveBuff - buff));
						WriteFile(dmp->dumpFile, dmp->string_location, wcslen((wchar_t*) dmp->string_location)*2, (LPDWORD) &writtenBytes, NULL);
					}

					wprintf(L"%s\r\n", moveBuff);
					WriteFile(dmp->dumpFile, moveBuff, wcslen((wchar_t*)moveBuff)*2, (LPDWORD) &writtenBytes, NULL);
					WriteFile(dmp->dumpFile, L"\r\n", 4, (LPDWORD) &writtenBytes, NULL);

					/*Skip the just found URL to find the next one*/
					moveBuff += wcslen((wchar_t*)moveBuff) -1;
				} else {
					moveBuff-=7;
				}
			}
		}
	} else {
		dmp->unlocated_size += memInfo->RegionSize;
	}
	return 0;
}

void DestroyFirefoxFilter1(DUMP_SESSION* filter) {
	Destroy_MemoryDump(filter);
}

/*-----------------------------------------------------End Filter 1-----------------------------------------------------*/


/*--------------------------------------------------------Filter 2--------------------------------------------------------*/

DUMP_SESSION* InitiateFirefoxFilter2(char* location, char* argument, int analyse_type) {
	return CreateMemoryDump(location, "Firefox_URL_Search_Methode2.txt", 0, 
		(argument==(char*) -1 ? DS_WRITE_LOCATION_STRING : strtol(argument, NULL, 10))
	);
}


/*Search for private visited URL's.*/
int ApplyFirefoxFilter2(DUMP_SESSION* dmp, HANDLE process, MEMORY_BASIC_INFORMATION* memInfo, char* buff, SIZE_T overflowSize) {
	/*In signature:

		"HTTP-memory-only-PB:"

	this is the string every URL starts with where somebody went in a private session.
	*/
	char signature[20]={0x48, 0x54, 0x54, 0x50, 0x2D, 0x6D, 0x65, 0x6D, 0x6F, 0x72, 0x79, 0x2D, 0x6F, 0x6E, 0x6C, 0x79, 0x2D, 0x50, 0x42, 0x3A};
	char* moveBuff=buff; //cur position in search
	int writtenBytes; //Returnvalue of Writefile()
	int readSize=0;  //temp paramenter for the size of the URL


	if(memInfo->State!=MEM_FREE && memInfo->State!=MEM_RESERVE) {
		for(; moveBuff<(buff + memInfo->RegionSize); moveBuff++) {
			if(memcmp(moveBuff, signature,  20) == 0) { //look for "HTTP-memory-only-PB:"
				/*write the address of the URL to the file*/
				if(dmp->flags&DS_WRITE_LOCATION_STRING) {
					printf("Address: %p: ", (memlong) memInfo->BaseAddress + (moveBuff - buff));
					sprintf(dmp->string_location, "Address: %p: ", (memlong) memInfo->BaseAddress + (moveBuff - buff));
					WriteFile(dmp->dumpFile, dmp->string_location, strlen(dmp->string_location), (LPDWORD) &writtenBytes, NULL);
				}

				moveBuff += 20; //skip "HTTP-memory-only-PB:"

				/*write the URL to the file*/
				printf("%s\n", moveBuff);
				WriteFile(dmp->dumpFile, moveBuff, strlen(moveBuff), (LPDWORD) &writtenBytes, NULL);
				WriteFile(dmp->dumpFile, "\r\n", strlen("\r\n"), (LPDWORD) &writtenBytes, NULL);

				//skip URL to find the next one
				moveBuff += strlen(moveBuff) -1;
			}
		}			
	} else {
		dmp->unlocated_size += memInfo->RegionSize;
	}
	return 0;
}


void DestroyFirefoxFilter2(DUMP_SESSION* filter) {
	Destroy_MemoryDump(filter);
}
/*-----------------------------------------------------End Filter 2-----------------------------------------------------*/

/*--------------------------------------------------------Filter 3--------------------------------------------------------*/

DUMP_SESSION* InitiateFirefoxFilter3(char* location, char* argument, int analyse_type) {
	return CreateMemoryDump(location, "Firefox_Cookie_Search.txt", 0, 
		(argument==(char*) -1 ? DS_WRITE_LOCATION_STRING : strtol(argument, NULL, 10))
	);
}

/*look for coockies*/
int ApplyFirefoxFilter3(DUMP_SESSION* dmp, HANDLE process, MEMORY_BASIC_INFORMATION* memInfo, char* buff, SIZE_T overflowSize) {
	int writtenBytes; //Returnvalue of Writefile()
	int isOkayURL; //Boolean: true if string is URL, false if not
	char* moveBuff=buff; //cur pos
	char* moveBuff2; //back-up cur pos

	long long validUntil=0; //Valid time for cookie in seconds since 1970
	struct tm* valid_Until_Struct;  //validUntil converted to a structure
	struct tm *loctime; //validUntil in current timezone

	char signature1[2]={0x98, 0xFE}; //Signature of a Cookie 19.0.2
	char signature2[2]={0x40, 0x07}; //Signature of a Cookie 21.0.0
	


	if(memInfo->State!=MEM_FREE && memInfo->State!=MEM_RESERVE) {
		for(; moveBuff<(buff + memInfo->RegionSize); moveBuff++) {
			if((
				memcmp(moveBuff, signature1, 2) == 0 //found 'signature?'
				|| memcmp(moveBuff, signature2, 2) == 0
				)
			&& overflowSize - ((memlong)moveBuff- (memlong)buff) > 0x40) { //header cookie is 0x40 bytes
				moveBuff2 = moveBuff;
				moveBuff2+=0x40;
				if(strlen(moveBuff2)!=0) { //is cookiename > 0 bytes
					moveBuff2+=strlen(moveBuff2)+1; //set pointer to value of cookie
					moveBuff2+=strlen(moveBuff2)+1; //set pointer to domain
					if(strlen(moveBuff2)!=0) { //is domain > 0 bytes

						/*Check if the domain meets the requirments:
						- letters a - z
						- numbers 0-9
						- - or . */
						isOkayURL = 1;
						while(*moveBuff2 != 0) {
							if((*moveBuff2 < 'a' || *moveBuff2 > 'z') 
								&& (*moveBuff2 < '0' || *moveBuff2 > '9') 
								&& *moveBuff2 != '-'
								&& *moveBuff2 != '.') {
									isOkayURL = 0;
									break;
							}
							moveBuff2++;
						}

						/*domainname check completed*/
							
						moveBuff2++; //set pointer to location where the cookie is valid
						if(*moveBuff2!='/') //the location starts with '/'
							isOkayURL=0;
							
						//All checks or ok? it is a cookie
						if(isOkayURL==1) { 

							//Read the valid date and convert to GMT and GMT+locale time
							valid_Until_Struct = (struct tm*) malloc(sizeof(struct tm));
							loctime =  (struct tm*) malloc(sizeof(struct tm));
							memcpy(&validUntil, moveBuff+0x20, 8);
							memcpy(valid_Until_Struct, gmtime((time_t*) validUntil), sizeof(struct tm));
							localtime_s (loctime, &validUntil);

							/*Read out and write to file:
							-Cookiename
							-Cookie value
							-valid date in vorm: hh:mm:ss dd-mm-yyyy
							-valid date in locale time
							-Domain where cookie is valid
							-URL where the cookie is valid
							*/

							if(dmp->flags&DS_WRITE_LOCATION_STRING) {
								printf("Address: %p: ", (memlong) memInfo->BaseAddress + (moveBuff - buff));
								sprintf(dmp->string_location, "Address: %p\r\n", (memlong) memInfo->BaseAddress + (moveBuff - buff));
								WriteFile(dmp->dumpFile, dmp->string_location, strlen(dmp->string_location), (LPDWORD) &writtenBytes, NULL);
							}

							moveBuff+=0x40;
							printf("Cookie: %s \n", moveBuff);//cookiename in console

							//cookiename
							WriteFile(dmp->dumpFile, "name: ", strlen("name: "), (LPDWORD) &writtenBytes, NULL); 
							WriteFile(dmp->dumpFile, moveBuff, strlen(moveBuff), (LPDWORD) &writtenBytes, NULL); 

							//cookie value
							WriteFile(dmp->dumpFile, "\r\nValue: ", strlen("\r\nvalue: "), (LPDWORD) &writtenBytes, NULL);
							moveBuff+=strlen(moveBuff)+1; //Zet pointer op cookie waarde
							WriteFile(dmp->dumpFile, moveBuff, strlen(moveBuff), (LPDWORD) &writtenBytes, NULL);

							//valid until
							WriteFile(dmp->dumpFile, "\r\nValid until: ", strlen("\r\nvalid until: "), (LPDWORD) &writtenBytes, NULL);

							if(validUntil==0x7FFFFFFFFFFFFFFF)
								sprintf(dmp->string_location, "Afther this session\r\n");
							else
								sprintf(dmp->string_location, "%hu:%hu:%hu %hu-%hu-%hu\r\n", 
									valid_Until_Struct->tm_hour,
									valid_Until_Struct->tm_min,
									valid_Until_Struct->tm_sec,
									valid_Until_Struct->tm_mday,
									valid_Until_Struct->tm_mon+1,
									valid_Until_Struct->tm_year + 1900);
							WriteFile(dmp->dumpFile, dmp->string_location, strlen(dmp->string_location), (LPDWORD) &writtenBytes, NULL);

							//Valid until (in local time)
							if(validUntil!=0x7FFFFFFFFFFFFFFF) {
								WriteFile(dmp->dumpFile, "\r\nValid until (local): ", strlen("\r\nValid until (local): "), (LPDWORD) &writtenBytes, NULL);

								sprintf(dmp->string_location, "%hu:%hu:%hu %hu-%hu-%hu\r\n", 
										loctime->tm_hour,
										loctime->tm_min,
										loctime->tm_sec,
										loctime->tm_mday,
										loctime->tm_mon+1,
										loctime->tm_year + 1900);

								WriteFile(dmp->dumpFile, dmp->string_location, strlen(dmp->string_location), (LPDWORD) &writtenBytes, NULL);
							}

							//domainname
							moveBuff+=strlen(moveBuff)+1;
							WriteFile(dmp->dumpFile, moveBuff, strlen(moveBuff), (LPDWORD) &writtenBytes, NULL);
							WriteFile(dmp->dumpFile, "\r\n", strlen("\r\n"), (LPDWORD) &writtenBytes, NULL);

							//location
							moveBuff+=strlen(moveBuff)+1;
							WriteFile(dmp->dumpFile, moveBuff, strlen(moveBuff), (LPDWORD) &writtenBytes, NULL);
							WriteFile(dmp->dumpFile, "\r\n\r\n", strlen("\r\n\r\n"), (LPDWORD) &writtenBytes, NULL);

							//free memory
							free(valid_Until_Struct);
							free(loctime); 
						}
					}
				}
			}
		}				
	} else {
		//if memory is not readable, nothing should happen...
	}

	return 0;
}


void DestroyFirefoxFilter3(DUMP_SESSION* filter) {
	Destroy_MemoryDump(filter);
}
/*-----------------------------------------------------End Filter 3-----------------------------------------------------*/